/*
 * PROJECT: Indiana University Stiquito Colony Simulator and Control Software
 *
 * (c) Copyright 1996 Jason L. Almeter, Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by Indiana University '' in the documentation or other
 * materials provided with the distribution and in all advertising
 * materials mentioning features or use of this software. The
 * name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * MODULE:
 *	Graphic display of colony/simulation virtual space
 *
 * FILE:
 * 	display.c
 *
 * ABSTRACT:
 * 	Display the locations of any stiquitos and pheromones in the colony
 *
 * $Revision: 1.4 $
 * $Date: 1996/01/19 14:28:50 $
 * $Author: jla $
 *
 * HISTORY:
 *
 * $Log: display.c,v $
 * Revision 1.4  1996/01/19  14:28:50  jla
 * #ifdef'ed out unstable code for release.
 *
 * Revision 1.3  1996/01/17  19:12:23  jla
 * Updated Cogyright verbage for release.
 *
 * Revision 1.2  1996/01/15  15:53:54  jla
 * Cleaning up the debugging code.
 *
 * Revision 1.1  1996/01/12  21:30:15  jla
 * Initial revision
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#ifdef SGI

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdlib.h>
#include "aux.h"

#else /* SPARC */

#include <GL/gl.h>
#include <GL/glu.h>
#include <stdlib.h>
#include "glaux.h"

#endif /* SGI */

#include "tca.h"
#include "../GPS/gps.h"
#include "../Motor/motorsim.h"
#include "../Clock/clock.h"

#define MODULE	"Display Module"

#define PI	3.1415926535897
#define ToDeg	(180.0/PI)
#define CIRCLE_POINTS 61		/* the number of points to use in
					 * drawing stiquito bounding circles
					 */

#define LENGTH	(15.0/18.0)		/* robot length relative to bounding
					 * radius */
#define LEG	(6.0/9.0)		/* relative leg length */
#define SHOULDER (5.0/9.0)		/* relative shoulder position */

int time;				/* current virtual time */

struct {				/* precomputed points for bounding */
   double x,z;				/* circles */
} circle[CIRCLE_POINTS];
		

struct {			  	/* latest robot positions */
   int flag;
   double x,z,theta;
} loc[MAX_ROBOTS];

/*static GLfloat spin1,spin2,spin3;	/**/
GLfloat fractX;/**/
GLfloat fractY;/**/
GLfloat fractZ;/**/
GLfloat fractDX;/**/
GLfloat fractDY;/**/
GLfloat fractDZ;/**/

CoordType robot[MAX_ROBOTS];		/* incoming update of robot positions */

/*PheromonePtr pher[N_COLORS];		/* linked lists of pheromones */


#define	P_COLORS	6	/* the maximum number of displayable pheromones
				 * this must be the number of entries in
				 * the table below */

GLfloat pherColors[][3] = {{0.0, 0.0, 1.0},	/* blue */
			   {1.0, 1.0, 0.0},	/* yellow */
			   {1.0, 0.0, 1.0},	/* magenta */
			   {0.0, 1.0, 1.0},	/* cyan */
			   {0.0, 0.0, 0.0},	/* black */
			   {1.0, 1.0, 1.0}};	/* white */


/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void clockHnd(TCA_REF_PTR ref, ClockPtr t)
 *
 * DESCRIPTION: 
 * 	update the virtual clock
 *
\*--------------------------------------------------------------------------*/

void clockHnd(TCA_REF_PTR ref, ClockPtr t)
{
   time = *t;

   tcaFreeData(tcaReferenceName(ref), t);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void gpsHnd(TCA_REF_PTR ref, CoordArrayPtr c)
 *
 * DESCRIPTION: 
 * 	update the current robot positions with the latest info from the
 * 	GPS system
 *
\*--------------------------------------------------------------------------*/

   
void gpsHnd(TCA_REF_PTR ref, CoordArrayPtr c)
{
   int i;
   double theta;

   if(c->n > MAX_ROBOTS) {
      fprintf(stderr, "GPS system update too many robots.\n");
      exit(-1);
   }

   
   for(i=0; i < c->n ; i++) {
      theta = robot[i].theta;
      robot[i].x = c->list[i].x;
      robot[i].y = c->list[i].y;
      robot[i].theta = c->list[i].theta;

      /* if the new theta is too far out of wack, the robot must have moved
       * backwards.  So tweak theta by PI
       */
      if(fabs(theta - robot[i].theta) < 0.5*PI ||
	 fabs(theta - robot[i].theta) > 1.5*PI) {
	 /* nop */
      } else {
	 robot[i].theta += PI;
	 if(robot[i].theta < PI)
	    robot[i].theta += 2.0*PI;
	 if(robot[i].theta > PI)
	    robot[i].theta -= 2.0*PI;
      }

   }

   tcaFreeData(tcaReferenceName(ref), c);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void pheromoneHnd(TCA_REF_PTR ref, PheromonePtr list);
 *
 * DESCRIPTION:
 * 	read a list of new pheromones from the pheromone manager
 * 	also expires any old pheromones
 *		(this should happen with the clock)
 *
\*--------------------------------------------------------------------------*/
#ifdef 0
void pheromoneHnd(TCA_REF_PTR ref, PheromonePtr c)
{
   int i,count;
   PheromonePtr temp,new;
 

   /* this really aught to return an error */
   if(!c) {
      tcaSuccess(ref);
      return;
   }

   /* add the new pheromones */
   temp = c;
   while(temp) {
      if(temp->color < 0 || temp->color >=N_COLORS) {
	 fprintf(stderr, "color %d out of range\n");
/*	 tcaFailure(ref, "color out of range", NULL);/**/
	 exit(-1);
      }
      if(!(new = (PheromonePtr)malloc(sizeof(PheromoneT)))) {
	 fprintf(stderr, "malloc failed\n");
	 exit(-1);
      }
      new->x = temp->x;
      new->y = temp->y;
      new->color = temp->color;
      new->expires = temp->expires;
      new->next = pher[new->color];
      pher[new->color] = new;
      temp = temp->next;
   }

/*   tcaSuccess(ref);/**/
   tcaFreeData(tcaReferenceName(ref), c);/**/

   /* expire any old pheromones		*/
   /* this should be in clockHnd	*/
   count = 0;
   for(i=0; i<N_COLORS; i++) {
      temp = pher[i];
      while(temp && time > temp->expires) {
	 printf("expire.\n");/**/
	 pher[i] = temp->next;
	 free(temp);
	 temp = pher[i];
      }
      while(temp) {
	 PheromonePtr old;
	 if(temp->next) {
	    if(time > temp->next->expires) {
	       printf("Expire.\n");
	       old = temp->next;
	       temp->next = temp->next->next;
	       free(old);
	       count++;
	    } else {
	       count++;
	       temp = temp->next;
	    }
	 } else {
	    temp = NULL;
	 count++;
	 }
      }
   }
   printf("%d\n",count);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void pher_display();
 *
 * DESCRIPTION:
 *	display the pheromone lists
 *
\*--------------------------------------------------------------------------*/

void pher_display()
{
   int i;
   PheromonePtr temp;

   glPushMatrix();
   glTranslatef(0.0, 0.250, 0.0);
   glRotatef(-90.0, 01.0, 0.0, 0.0);
   glPointSize(6.0);

   glBegin(GL_POINTS);

   for(i=0; i<N_COLORS; i++) {
      temp = pher[i];
      while(temp) {
	 glColor3fv(pherColors[temp->color]);/**/
	 glVertex2f(temp->y, -temp->x);
	 temp = temp->next;
      }
   }
   glEnd();
   glPopMatrix();
}
#endif

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void stiq_display();
 *
 * DESCRIPTION:
 * 	draw a stiquito
 *
\*--------------------------------------------------------------------------*/

void stiq_display()
{
   int i;
   
   glColor3f(0.0, 1.0, 0.0);/**/

   glPushMatrix();
   glTranslatef(0.0, 0.50, 0.0);
   glRotatef(-90.0, 01.0, 0.0, 0.0);
/*   glRectf(-1.0, ROBOT_RADIUS, 1.0, -ROBOT_RADIUS);/**/
   glRectf(-LENGTH*ROBOT_RADIUS, 1.0, LENGTH*ROBOT_RADIUS,-1.0);
   glBegin(GL_LINE_LOOP);
   for(i=0; i<CIRCLE_POINTS; i++)
      glVertex2f(circle[i].x, circle[i].z);
   glEnd();
   glBegin(GL_LINES);
   glVertex2f(0.0, -LEG*ROBOT_RADIUS);
   glVertex2f(0.0, LEG*ROBOT_RADIUS);
   glVertex2f(SHOULDER*ROBOT_RADIUS, -LEG*ROBOT_RADIUS);
   glVertex2f(SHOULDER*ROBOT_RADIUS, LEG*ROBOT_RADIUS);
   glVertex2f(-SHOULDER*ROBOT_RADIUS, -LEG*ROBOT_RADIUS);
   glVertex2f(-SHOULDER*ROBOT_RADIUS, LEG*ROBOT_RADIUS);
   glEnd();
   glPointSize(6.0);
   glColor3f(01.0, 0.0, 0.0);/**/
   glBegin(GL_POINTS);/**/
   glVertex2f(-ROBOT_RADIUS*0.95,0.0);/**/
   glEnd();/**/
   glPopMatrix();

}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void sim_display()
 *
 * DESCRIPTION:
 * 	draw the current state of the simulation/colony
 *
\*--------------------------------------------------------------------------*/

void sim_display()
{
   int i;
   char buf[10];

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	/* clear to start */

   glLoadIdentity ();

   gluLookAt(fractX, fractY, fractZ,	/* where is the camera */
	     0.0, 0.5, 0.0,	/* where is it pointed */
	     1.0, 0.0, 0.0);	/* what is up */

   /* draw the things *********************************/

   glPushMatrix();
   glPushMatrix();

   glRotatef(-90.0, 1.0, 0.0, 0.0);

   glColor3f(0.70, 0.70, 0.70);/**/

   /* this may need to be adjusted */
   glRectf(61.0,61.0,-61.0,-61.0);
   glPopMatrix();
   /* camera orbiting origin */

   /*   pher_display();/**/

   for(i=0; i<MAX_ROBOTS; i++) {
      sprintf(buf,"%d",i);
      glPushMatrix();
      glTranslatef(robot[i].y, 0.0, robot[i].x);
      glPushMatrix();
      glTranslatef(0.0,2.5,0.0);
      glRotatef(180.0,1.0,0.0,0.0);
      glColor3f(1.0,0.0,0.0);
      glRasterPos3f(0.7*ROBOT_RADIUS,0.0,-0.7*ROBOT_RADIUS);
      printString(buf);
      glPopMatrix();
      glRotatef(90.0+ToDeg*robot[i].theta, 0.0, 1.0, 0.0);/**/
      stiq_display();
      glPopMatrix();
   }
   glPopMatrix();

   glFlush();		/* force the draw */

#ifdef SGI
   glXSwapBuffers (auxXDisplay(), auxXWindow());
#else
   tkSwapBuffers();	/* display the buffer */
#endif /* SGI */
}	

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION:
 *	void sim_update()
 *
 * DESCRIPTION:
 *	update structures for next anim step
 *
\*--------------------------------------------------------------------------*/


void sim_update()
{
   int number;
   CoordType coord;
   int status;

   while(Success == tcaHandleMessage(0));
}	

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void sim_init()
 *
 * DESCRIPTION:
 * 	initialize the simulator
 *
\*--------------------------------------------------------------------------*/


void sim_init()
{
   int i;
  GLfloat norm;

   time = 0;		/* virtual time */
   
   /*   if(P_COLORS < N_COLORS) {
	/*      fprintf(stderr, "Not enough colors for pheromones.\n");
		/*      exit(-1);
			/*   }/**/
   

   for(i=0; i<MAX_ROBOTS; i++) {
/*      robot[i].x = EAST_WALL*2.0;/**/
      robot[i].x = 0.0;
      robot[i].y = NORTH_WALL*2.0;
      robot[i].theta = 0.5*PI;
   }

   /* precompute points for bounding circle */
   for(i=0; i<CIRCLE_POINTS; i++) {
      circle[i].x = ROBOT_RADIUS*cos(2.0*PI*i/CIRCLE_POINTS);
      circle[i].z = ROBOT_RADIUS*sin(2.0*PI*i/CIRCLE_POINTS);
   }

   /* no pheromones to start */
   /*   for(i=0; i<N_COLORS; i++)
	/*      pher[i] = NULL;/**/
   
   fractX = 0.0;
   fractY = 200.0;
   fractZ = 0.0;
   fractDX = 0.0;
   fractDY = -1.5;
   fractDZ = -8.0;

   norm = (GLfloat)sqrt(fractDX*fractDX+fractDY*fractDY+fractDZ*fractDZ);
   fractDX /= norm;
   fractDY /= norm;
   fractDZ /= norm;

   printf("Connect...\n");
   tcaConnectModule(MODULE, tcaServerMachine());

   tcaWaitUntilReady();

   tcaRegisterHandler(CoordBroadcast, "gpsHnd", gpsHnd);
   tcaLimitPendingMessages(CoordBroadcast, MODULE, 1);/**/

   /*   tcaRegisterInformMessage(PheromoneInform, PheromoneFormat);
	/*   tcaRegisterHandler(PheromoneInform, "pheromoneHnd", pheromoneHnd);
	     /**/

   tcaRegisterHandler(ClockBroadcast, "clockHnd", clockHnd);
   tcaLimitPendingMessages(ClockBroadcast, MODULE, 1);
}



