/*
 * PROJECT: Indiana University Stiquito Colony Simulator and Control Software
 *
 * (c) Copyright 1996 Jason L. Almeter, Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by Indiana University '' in the documentation or other
 * materials provided with the distribution and in all advertising
 * materials mentioning features or use of this software. The
 * name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * MODULE:
 *	Position sensor system for colony.
 *
 * FILE:
 *	gps.c
 *
 * ABSTRACT:
 *	receives updates from vision system and integrates into robot
 *	locations.
 *
 * $Revision: 1.4 $
 * $Date: 1996/01/17 18:59:10 $
 * $Author: jla $
 *
 * HISTORY:
 *
 * $Log: gps.c,v $
 * Revision 1.4  1996/01/17  18:59:10  jla
 * Updated Copyright verbage in preparation for release.
 *
 * Revision 1.3  1996/01/16  20:02:50  jla
 * numberQueryHnd returns an error for out of range now.
 *
 * Revision 1.2  1996/01/15  13:50:31  jla
 * Cleaning up some of the residual mess.
 *
 * Perhaps it should run from the virtual clock instead of being driven
 * by the vision system?
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "tca.h"
#include "gps.h"
#include "../Motor/motor.h"

#define MODULE	"GPS Module"

int robot;		/* currently allocated robots 		*/
CoordArray coords;	/* robot coordinates for transmission	*/

CoordType reply;	/* structure for answering a coord query
			 * no longer used?			*/


/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void numberQueryHnd(TCA_REF_PTR ref, IntPtr number);
 *
 * DESCRIPTION:
 *	Allocates a 'body' to a controller.
 *	This may be moved to the motor controller in the future and
 *	gain parameters for access control.
 *
\*--------------------------------------------------------------------------*/

void numberQueryHnd(TCA_REF_PTR ref, IntPtr number)
{
   if(robot >= MAX_ROBOTS) {
      tcaNullReply(ref);
      tcaFreeData(tcaReferenceName(ref), number);
      fprintf(stderr, "%d out of range.\n", robot);
      return;
   }
   printf("just met robot controller %d\n", robot);
   tcaReply(ref, &robot);
   tcaFreeData(tcaReferenceName(ref), number);
   robot++;
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void coordQueryHnd(TCA_REF_PTR ref, IntPtr number)
 *
 * DESCRIPTION: 
 * 	tell a controller where robot 'number' is
 *
\*--------------------------------------------------------------------------*/

void coordQueryHnd(TCA_REF_PTR ref, IntPtr number)
{
   if(*number >= coords.n || *number < 0) {
      tcaNullReply(ref);
      tcaFreeData(tcaReferenceName(ref), number);
      return;
   }

   printf("Sending coords of robot %d %f %f %e\n",*number,
	  coords.list[*number].x,
	  coords.list[*number].y,
	  coords.list[*number].theta);

   reply.x = coords.list[*number].x;
   reply.y = coords.list[*number].y;
   reply.theta = coords.list[*number].theta;

   tcaReply(ref, &reply);
   tcaFreeData(tcaReferenceName(ref), number);
}


/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void coordInformHnd(TCA_REF_PTR ref, CoordArrayPtr c)
 *
 * DESCRIPTION: 
 * 	received a list of blob coords from vision; deal with it
 *
\*--------------------------------------------------------------------------*/

void coordInformHnd(TCA_REF_PTR ref, CoordArrayPtr c)
{
   int i,j;
   double dx,dy;

   coordGrowArray(c->n, &coords);

   /* ### this should test for theta out of whack, and add PI for reverse
      travel */
   for(j=0; j<coords.n; j++) {
      dx = c->list[j].x - coords.list[j].x;
      dy = c->list[j].y - coords.list[j].y;
      if(dx != 0.0 || dy != 0.0) {
	 coords.list[j].theta = atan2(dy,dx);
      } 
      coords.list[j].x = c->list[j].x;
      coords.list[j].y = c->list[j].y;
   }
   tcaFreeData(tcaReferenceName(ref), c);

   tcaBroadcast(CoordBroadcast, &coords);/**/
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: main()
 *
 * DESCRIPTION: 
 *
\*--------------------------------------------------------------------------*/


main()
{
   robot = 0;

   coordInitArray(&coords);

   printf("Connect...\n");

   tcaConnectModule(MODULE, tcaServerMachine());

   tcaRegisterBroadcastMessage(CoordBroadcast, CoordBroadcastFormat);

   tcaRegisterInformMessage(CoordInform, CoordInformFormat);
   tcaRegisterHandler(CoordInform, "coordInformHnd", coordInformHnd);

/*   tcaRegisterQueryMessage(CoordQuery, "int", CoordQueryFormat);
/*   tcaRegisterHandler(CoordQuery, "coordQueryHnd", coordQueryHnd);/**/

   tcaRegisterQueryMessage(NumberQuery, NULL, IntFormat);
   tcaRegisterHandler(NumberQuery, "numberQueryHnd", numberQueryHnd);/**/

   tcaWaitUntilReady();

   tcaModuleListen();   /**/

}
